/*
 * Eigen values and eigen vectors of matrix
 * written by KT (ref. CHEQRS in NUMPAC)
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "complex.h"
#include "cmatutil.h"
#include "vecmat.h"

#define sq(x)	((x)*(x))
#define MAX2(x, y) ((x) > (y) ? (x) : (y))
#define EPS_E 1e-13		/* Recommended value for eps */
#define ITER_MAX 30		/* Iteration limit */

void print_complexmat(int n, complex **a);
void re_part_csort(int n, complex *eigen_val, complex **eigen_vec, int ind);
void norm_csort(int n, complex *eigen_val, complex **eigen_vec, int ind);
int eigen_c_mat(  int n,  double eps, int ind,
                  complex **za, complex *eigen_val, complex **eigen_vec );


int eigen_c_mat(
  int n,  double eps, int ind,
  complex **a, complex *eigen_val, complex **eigen_vec
  )
     /*
      * Eigen values and eigen vectors of complex general matrix
      *  using the QR method
      *  written by KT (ref. CHEQRS in NUMPAC)
      *
      * input: ind  =  1: sort eigenvalues in the descending order by the Re-part of eigen_val
      *        ind  = -1: sort eigenvalues in the ascending order  by the Re-part of eigen_val
      *        ind >=  2: sort eigenvalues in the descending order by the norm of eigen_val
      *        ind <= -2: sort eigenvalues in the ascending order  by the norm of eigen_val
      *        input value of matrix a will be changed.
      * return:
      *        0: normal.
      *        1: n=1 or a is a zero matrix.
      *        2: no convergence
      *        3: invalid input parameter
      *        input matrix a is changed.
      */
{
  int i,j,k,l,m;
  int ll,lu,it;
  int *iw;
  double a_norm, del;
  double am,aa;
  double r;
  complex z;
  complex zz,zx,zy,zc,zcc,zcs,zs;
  
  iw = new_vector(int, n);		chk_vector(iw);
  
  if( (n < 1) || (eps <= 0.0) ){
    fprintf(stderr,"$$ (eigen_c_mat) invalid n or epsilon.\n");
    free_vector(iw);
    return 3;
  }
  
  /*
	  Compute the matrix norm with respect to the magnitude
	  and set the convergence criterion.
   */
  a_norm = 0.0;
  for(j=0; j<n; j++){
    iw[j] = n;
    r = 0.0;
    for(i=0; i<n; i++){ r += c_abs( a[i][j] ); }
    a_norm = MAX2( a_norm, r );
  }
  if ( n == 1 || fabs(a_norm) < eps ){
    /* n is 1 or a is a zero matrix. */
    for(j=0; j<n; j++){
      for(i=0; i<n; i++){ eigen_vec[i][j] = c_conv(0.0, 0.0); }
      eigen_vec[j][j] = c_conv(1.0, 0.0);
      eigen_val[j] = a[j][j];
    }
    free_vector(iw); /* free_matrix(a); */ return 1;
  }
  del = MAX2(eps, EPS_E) * a_norm;	/* <<<<< */
  /* printf("del = %lg\n", del); */
  
  if ( n != 2 ){
    /*
		  Stabilized elementary transformation to the Hessenberg form.
		  iw[i] holds the row number of the element which is brought
		  to the pivot position (i+1,i) in the i-th stage.
		  When iw[i] holds n, it means that no transformation is
		  performed in the i-th stage. 
     */
    for(l=0; l<n-2; l++){
      m = l + 1;
      am = c_abs( a[l+1][l] );
      for(i=l+2; i<n; i++){
        aa = c_abs( a[i][l] );
        if (aa > am){ m = i; am = aa; }
      }
      if (am > del){
        iw[l] = m; z = a[m][l];
        if (m != l+1){ a[m][l] = a[l+1][l];	a[l+1][l] = z; }
        /*
				  Left-hand transformation (Gaussian elimination
				  with the element a[l+1][l] as the pivot).
         */
        z = c_div( c_conv(-1.0,0.0), z );
        for(i=l+2; i<n; i++){ a[i][l] = c_mul( a[i][l], z ); }
        for(j=l+1; j<n; j++){
          z = a[m][j];
          if (m != l+1){
            a[m][j] = a[l+1][j];	a[l+1][j] = z;
          }
          for(i=l+2; i<n; i++){
            a[i][j] = c_add( c_mul(a[i][l],z), a[i][j] );
          }
        }
        /*	Right-hand transformation. */
        if(m != l+1){
          for(i=0; i<n; i++){
            z = a[i][m];	a[i][m] = a[i][l+1];	a[i][l+1] = z;
          }
        }
        for(i=0; i<n; i++){
          z = c_conv(0.0,0.0);
          for(k = l+2; k<n; k++){ z = c_add( c_mul(a[i][k], a[k][l]), z ); }
          a[i][l+1] = c_sub( a[i][l+1], z );
        }
      }
    }
  }
  
  if( ind != 0 ){ /*  Accumulate stabilized elementary transformations in eigen_vec. */
    for(j=0; j<n; j++){
      for(i=0; i<n; i++){
        eigen_vec[i][j] = c_conv(0.0,0.0);
      }
      if (j != 0 && j != n-1){
        for(i=j+1; i<n; i++){
          eigen_vec[i][j] = c_mul( c_conv(-1.0,0.0), a[i][j-1] );
        }
      }
      eigen_vec[j][j] = c_conv(1.0, 0.0);
    }
    if( n != 2 ){
      for(i=n-2; i>0; i--){
        m = iw[i-1];
        if (m != n && m != i){
          for(j=i; j<n; j++){
            zz = eigen_vec[i][j];
            eigen_vec[i][j] = eigen_vec[m][j];
            eigen_vec[m][j] = zz;
          }
        }
      }
    }
  }
  
  /*
      QR-transformation of the complex upper Hessenberg matrix.
      ll and lu hold the lower and upper limit of the current
      submatrix under treatment.
   */
  for(lu = n - 1; lu > 0;	lu--){
    it = 0;
    
    /* Search for a small subdiagonal element. */
    for(;;){
      for(ll = lu; ll > 0; ll--){
        if ( c_abs(a[ll][ll-1]) <= del ){ break; }
      }
      
      if(ll == lu){ break; } /* Loop exit */ 
      
      /*
			 If convergence is not attained within 30 iterations,
			 trancate the calculation.
       */
      if(it > ITER_MAX){
        free_vector(iw); /* free_matrix(a); */ return 30000;
      }
      /*
			  Determine the origin shift as the eigenvalue of the
			  2 by 2 bottom submatrix which is nearer to the
			  bottom element.
       */
      zx = c_div( c_add( a[lu-1][lu-1], a[lu][lu] ), c_conv(2.0,0.0) );
      zy = c_sub(zx, a[lu][lu]);
      zz = c_sqrt( c_add( c_mul( a[lu-1][lu], a[lu][lu-1] ), c_mul( zy, zy ) ) );
      
      if ( c_abs( c_sub(zy,zz) ) - c_abs( c_add(zy,zz) ) < 0.0 ){
        zz = c_sub(zx, zz);
      }
      else{
        zz = c_add(zx, zz);
      }
      
      a[ll][ll] = c_sub( a[ll][ll], zz );
      zc = c_conv(1.0,0.0);
      
      for(i=ll;i<=lu;i++){
        if(i == ll){ 	/* Left-hand transformation */
          zx = a[i][i];	zy = a[i+1][i];
          r = sqrt( sq(zx.re) + sq(zx.im) + sq(zy.re) + sq(zy.im) );
          a[i][i] = c_mul( zc, c_conv(r,0.0) );
          if (i > ll){ a[i][i-1] = c_mul(zs, c_conv(r,0.0)); }
          a[i+1][i+1] = c_sub( a[i+1][i+1], zz );
          zcc = c_div( zx, c_conv(r,0.0) );
          zs = c_div(zy, c_conv(r,0.0) );
          zc = c_conj(zcc);
          zcs = c_conj(zs);
          for(j=i+1;j<n;j++){
            zy = a[i][j];
            a[i][j] = c_add( c_mul(zy,zc), c_mul(a[i+1][j], zcs) );
            a[i+1][j] = c_sub( c_mul( a[i+1][j],zcc ), c_mul( zy,zs ) );
          }
        }
        else{
          for(j=0;j<=i-1;j++){	/* Right-hand transformation */
            zy = a[j][i-1];
            a[j][i-1] = c_add( c_mul(zy,zcc), c_mul(a[j][i],zs) );
            a[j][i] = c_sub( c_mul(a[j][i],zc), c_mul(zy,zcs) );
          }
          a[i-1][i-1] = c_add( a[i-1][i-1], zz );
          
          if(ind != 0){	/* Accumulate right-hand transformations in eigen_vec. */
            for(j=0;j<n;j++){
              zy = eigen_vec[j][i-1];
              eigen_vec[j][i-1] = c_add( c_mul(zy,zcc), c_mul(eigen_vec[j][i],zs) );
              eigen_vec[j][i] = c_sub( c_mul(eigen_vec[j][i],zc), c_mul(zy,zcs) );
            }
          }
          if(i != lu){	/* Left-hand transformation */
            zx = a[i][i];
            zy = a[i+1][i];
            r = sqrt( sq(zx.re) + sq(zx.im) + sq(zy.re) + sq(zy.im) );
            a[i][i] = c_mul( zc, c_conv(r,0.0) );
            if (i > ll){ a[i][i-1] = c_mul(zs, c_conv(r,0.0)); }
            a[i+1][i+1] = c_sub( a[i+1][i+1], zz );
            zcc = c_div( zx, c_conv(r,0.0) );
            zs = c_div(zy, c_conv(r,0.0) );
            zc = c_conj(zcc);
            zcs = c_conj(zs);
            for(j=i+1;j<n;j++){
              zy = a[i][j];
              a[i][j] = c_add( c_mul(zy,zc), c_mul(a[i+1][j], zcs) );
              a[i+1][j] = c_sub( c_mul( a[i+1][j],zcc ), c_mul( zy,zs ) );
            }
          }
        }
      }
      a[lu][lu-1] = c_mul( a[lu][lu], zs );
      a[lu][lu] = c_add( c_mul(a[lu][lu], zc), zz );
      it++;	/* it: Iteration counter */
    }
  }
  
  /*
	  QR-transformation is over. Eigenvalues are given by the diagonal
	  elements of the resultant upper triangular matrix.
   */
  
  for( i=0; i<n; i++ ){ eigen_val[i] = a[i][i]; }
  if (ind == 0){
    free_vector(iw); /* free_matrix(a);*/ return 0;		/* Normal exit. */
  }
  
  /*
		Determine the eigenvectors of the resultant upper triangular
		matrix and transform them back to those of the original
		matrix.
   */
  for(i=0; i<n; i++){ a[i][i] = c_conv(1.0,0.0); }
  for(j=n-1; j>0; j--){
    for(i=j-1; i>=0; i--){
      zs = c_conv(0.0,0.0);
      for(k=i+1; k<=j; k++){ zs = c_add( c_mul( a[i][k], a[k][j] ), zs ); }
      zz = c_sub( eigen_val[j], eigen_val[i] );
      if(c_abs(zz) < del){ zz = c_conv(del,0.0); }
      a[i][j] = c_div(zs, zz);
    }
    for(i=0; i<n; i++){
      zs = c_conv(0.0,0.0);
      for(k=0; k<=j; k++){ zs = c_add( c_mul( eigen_vec[i][k], a[k][j] ), zs ); }
      eigen_vec[i][j] = zs;
    }
  }
  
  /* Normalize eigenvectors with respect to norm. */
  for(j=0; j<n; j++){
    r = 0.0;
    for(i=0; i<n; i++){
      r += sq(eigen_vec[i][j].re) + sq(eigen_vec[i][j].im);
    }
    r = 1.0 / sqrt(r);
    for(i=0; i<n; i++){
      eigen_vec[i][j] = c_mul( eigen_vec[i][j], c_conv(r,0) );
    }
  }
  /*
	for(i=0; i<n; i++){
		printf("$$ eigenval[%d] = %lf %lf, eigen_vec[0][%d]^2 = %lf\n",
			i, eigen_val[i].re, eigen_val[i].im,
			i, sq(eigen_vec[0][i].re) + sq(eigen_vec[0][i].im) );
	}
   */
  
  /* Sort eigenvalues and eigenvectors */
  if(ind == 1 || ind == -1){	/* according to the Re-part of eigen values */
    re_part_csort(n, eigen_val, eigen_vec, ind);
  }
  else{						/* according to the norm of eigen values */
    norm_csort(n, eigen_val, eigen_vec, ind);
  }
  free_vector(iw); /* free_matrix(a); */
  return 0;		/* Normal exit. */
}


void norm_csort(int n, complex *eigen_val, complex **eigen_vec, int ind)
{
  double aa, al, s;
  int i, j, l, ii, ll;
  complex z;
  
  /* Sort eigenvalues and eigenvectors in the decreasing order of
		the norm of eigenvalues. */
  s = 1.0;
  if ( ind < 0 ) { s = -s; }
  
  j = n - 1;		/* <<<<< */
  do{
    l = 0; ii = 0; ll = 0;
    al = ( sq(eigen_val[0].re) + sq(eigen_val[0].im) ) * s;	/* <<<<< */
    for(i=1;i<=j;i++){			/* <<<<< */
      aa = ( sq(eigen_val[i].re) + sq(eigen_val[i].im) ) * s;
      if(aa <= al){ l = i;  al = aa; }
      else{ ll = l;  ii = i; }
    }
    if(ii != ll){
      z = eigen_val[ll];
      eigen_val[ll] = eigen_val[ii];
      eigen_val[ii] = z;
      for(i=0;i<n;i++){
        z = eigen_vec[i][ll];
        eigen_vec[i][ll] = eigen_vec[i][ii];
        eigen_vec[i][ii] = z;
      }
    }
    j = ii - 1;
  } while(j >= 1);
}


void re_part_csort(int n, complex *eigen_val, complex **eigen_vec, int ind)
{
  double aa, al, s;
  int i, j, l, ii, ll;
  complex z;
  
  /* Sort eigenvalues and eigenvectors in the decreasing order of
		the real part of eigenvalues. */
  
  s = 1.0;
  if ( ind < 0 ) { s = -s; }
  
  j = n - 1;		/* <<<<< */
  do{
    l = 0; ii = 0; ll = 0;
    al = eigen_val[0].re * s;		/* <<<<< */
    for(i=1; i<=j; i++){				/* <<<<< */
      aa = eigen_val[i].re * s;
      if(aa <= al){ l = i;  al = aa; }
      else{ ll = l;  ii = i; }
    }
    if(ii != ll){
      z = eigen_val[ll];
      eigen_val[ll] = eigen_val[ii];
      eigen_val[ii] = z;
      for(i=0;i<n;i++){
        z = eigen_vec[i][ll];
        eigen_vec[i][ll] = eigen_vec[i][ii];
        eigen_vec[i][ii] = z;
      }
    }
    j = ii - 1;
  } while(j >= 1);
}
