/*
 * lattice.c
 *   crystallographic functions by KT
 *   Apl. 1995
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "lattice.h"

#define	sq(x)	( (x)*(x) )
#define	PI	3.141592653589793
/*
#define YES 1
#define NO 0
 */
#define EPS 1e-12


CELL_PAR bravai(LAT_PAR lp)
     /*
      * calc. of bv (which is needed for calc. of proi)
      *  input: a,b,c,alpha,beta,gamma (angstrom ,degree)
      */
{
  double alp, bet, gam, as, bs, cs;
  double sin_alp, sin_bet, sin_gam, cos_alp, cos_bet, cos_gam;
  double cos_alps, cos_bets, cos_gams;
  CELL_PAR bv;
  
  alp = lp.alpha * PI / 180.0;		/* degree to radian */
  bet = lp.beta  * PI / 180.0;
  gam = lp.gamma * PI / 180.0;
  sin_alp = sin(alp);	sin_bet = sin(bet);	sin_gam = sin(gam);
  cos_alp = cos(alp);	cos_bet = cos(bet);	cos_gam = cos(gam);
  
  bv.s11 = sq(lp.b) * sq(lp.c) * sq(sin_alp);
  bv.s22 = sq(lp.a) * sq(lp.c) * sq(sin_bet);
  bv.s33 = sq(lp.a) * sq(lp.b) * sq(sin_gam);
  bv.s12 = lp.a * lp.b * sq(lp.c) * ( cos_alp * cos_bet - cos_gam );
  bv.s23 = sq(lp.a) * lp.b * lp.c * ( cos_bet * cos_gam - cos_alp );
  bv.s13 = lp.a * sq(lp.b) * lp.c * ( cos_gam * cos_alp - cos_bet );
  
  /* vc: Cell volume ( = sqrt(det(G)) ) */
  bv.vc = lp.a * lp.b * lp.c
    * sqrt(
      1.0 - sq(cos_alp) - sq(cos_bet) - sq(cos_gam)
      + 2.0 * cos_alp * cos_bet * cos_gam
      );
  
  /* Metric tensor G */
  bv.aa = sq(lp.a);
  bv.bb = sq(lp.b);
  bv.cc = sq(lp.c);
  bv.ab = lp.a * lp.b * cos_gam;
  bv.bc = lp.b * lp.c * cos_alp;
  bv.ac = lp.a * lp.c * cos_bet;
  
  /* Metric tensor G_inv */
  as = lp.b * lp.c * sin_alp / bv.vc;	/* a_star */
  bs = lp.c * lp.a * sin_bet / bv.vc;	/* b_star */
  cs = lp.a * lp.b * sin_gam / bv.vc;	/* c_star */
  cos_alps = (cos_bet * cos_gam - cos_alp) / (sin_bet * sin_gam);	/* cos(alpha_star) */
  cos_bets = (cos_gam * cos_alp - cos_bet) / (sin_gam * sin_alp);	/* cos(beta_star ) */	
  cos_gams = (cos_alp * cos_bet - cos_gam) / (sin_alp * sin_bet);	/* cos(gamma_star) */
  bv.aas = sq(as);
  bv.bbs = sq(bs);
  bv.ccs = sq(cs);
  bv.abs = as * bs * cos_gams;
  bv.bcs = bs * cs * cos_alps;
  bv.acs = as * cs * cos_bets;
  
  return bv;
}


double inner_product(R_HKL g1, R_HKL g2, CELL_PAR bv)
     /***********************************************
 inner product of reciprocal vectors g1 and g2
      ***********************************************/
{
  /*
	return
	( bv.s11 * g1.h * g2.h
	+ bv.s22 * g1.k * g2.k
	+ bv.s33 * g1.l * g2.l
	+ bv.s23 * ( g1.k * g2.l + g2.k * g1.l )
	+ bv.s13 * ( g1.l * g2.h + g2.l * g1.h )
	+ bv.s12 * ( g1.h * g2.k + g2.h * g1.k )
	) / sq(bv.vc);
   */
  
  /* = g1^t * G_inv * g2 */
  return
    g1.h * g2.h * bv.aas
      + g1.k * g2.k * bv.bbs
        + g1.l * g2.l * bv.ccs
          + (g1.h * g2.k + g1.k * g2.h) * bv.abs
            + (g1.k * g2.l + g1.l * g2.k) * bv.bcs
              + (g1.l * g2.h + g1.h * g2.l) * bv.acs;
}


double sq_abs(R_HKL g, CELL_PAR bv)
     /*************************************************
 square of absolute value of a reciprocal vector g
      *************************************************/
{
  /* return inner_product(g,g,bv); */
  /*   = g^t * G_inv * g   */
  /*
	return 
		sq(g.h) * bv.aas + sq(g.k) * bv.bbs + sq(g.l) * bv.ccs
		 + 2 * ( g.h * g.k * bv.abs + g.k * g.l * bv.bcs + g.h * g.l * bv.acs );
   */
  return inner_product(g, g, bv);
}


R_HKL hkl2rhkl(HKL g)
{
  R_HKL rg;
  rg.h = (double) g.h;	rg.k = (double) g.k;	rg.l = (double) g.l;
  return rg;
}


double inner_product_uvw_hkl(UVW u, R_HKL g)
     /*
      * Inner product of real and reciprocal lattice vectors
      */
{
  return (double) u.u * g.h + (double) u.v * g.k + (double) u.w * g.l;
}


/* double inner_product_uvw(UVW u1, UVW u2, LAT_PAR lp) */
double inner_product_uvw(UVW u1, UVW u2, CELL_PAR bv)
     /*
      * Inner product of real-lattice vector u1 and u2
      */
{
  /*
	return
		  (double)( u1.u * u2.u ) * sq(lp.a)
		+ (double)( u1.v * u2.v ) * sq(lp.b)
		+ (double)( u1.w * u2.w ) * sq(lp.c)
		+ (double)( u1.u * u2.v + u1.v * u2.u )
   * lp.a * lp.b * cos(lp.gamma * PI / 180.0)
		+ (double)( u1.v * u2.w + u1.w * u2.v )
   * lp.b * lp.c * cos(lp.alpha * PI / 180.0)
		+ (double)( u1.w * u2.u + u1.u * u2.w )
   * lp.c * lp.a * cos(lp.beta  * PI / 180.0);
   */
  
  return
    (double)( u1.u * u2.u ) * bv.aa
      + (double)( u1.v * u2.v ) * bv.bb
        + (double)( u1.w * u2.w ) * bv.cc
          + (double)( u1.u * u2.v + u1.v * u2.u ) * bv.ab
            + (double)( u1.v * u2.w + u1.w * u2.v ) * bv.bc
              + (double)( u1.w * u2.u + u1.u * u2.w ) * bv.ac;
}


/* double abs_uvw(UVW u, LAT_PAR lp) */
double abs_uvw(UVW u, CELL_PAR bv)
     /*
      * Absolute value of real lattice vector u  [A]
      */
{
  /*
	return sqrt(
		  sq((double)u.u) * sq(lp.a)
		+ sq((double)u.v) * sq(lp.b)
		+ sq((double)u.w) * sq(lp.c)
		+ 2 * (
				(double)u.u * (double)u.v
   * lp.a * lp.b * cos(lp.gamma * PI / 180.0)
			  + (double)u.v * (double)u.w
   * lp.b * lp.c * cos(lp.alpha * PI / 180.0)
			  + (double)u.w * (double)u.u
   * lp.c * lp.a * cos(lp.beta  * PI / 180.0)
			  )
		);
   */
  return sqrt( inner_product_uvw(u, u, bv) );
}


R_UVW uvw2ruvw(UVW u)
{
  R_UVW ru;
  ru.u = (double) u.u;	ru.v = (double) u.v;	ru.w = (double) u.w;
  return ru;
}


double inner_product_r_uvw(R_UVW u1, R_UVW u2, CELL_PAR bv)
     /*
      * Inner product of real-lattice vector u1 and u2
      */
{
  return
    u1.u * u2.u * bv.aa
      + u1.v * u2.v * bv.bb
        + u1.w * u2.w * bv.cc
          + ( u1.u * u2.v + u1.v * u2.u ) * bv.ab
            + ( u1.v * u2.w + u1.w * u2.v ) * bv.bc
              + ( u1.w * u2.u + u1.u * u2.w ) * bv.ac;
}


/* double abs_r_uvw(R_UVW u, LAT_PAR lp) */
double abs_r_uvw(R_UVW u, CELL_PAR bv)
     /*
      * Absolute value of real lattice vector u  [A]
      */
{
  /*	return sqrt(
		  sq(u.u) * sq(lp.a)
		+ sq(u.v) * sq(lp.b)
		+ sq(u.w) * sq(lp.c)
		+ 2 * (
				u.u * u.v * lp.a * lp.b * cos(lp.gamma * PI / 180.0)
			  + u.v * u.w * lp.b * lp.c * cos(lp.alpha * PI / 180.0)
			  + u.w * u.u * lp.c * lp.a * cos(lp.beta  * PI / 180.0)
			  )
		);
   */
  return sqrt( inner_product_r_uvw(u, u, bv) );
}


R_HKL zolz_ort_base(UVW u, R_HKL g1, CELL_PAR bv)
     /***************************************
 calc. of zolz orthogonal base vectors
 input : u (real space)
         g1(reciprocal space)
 return : vector
  perpendicular to u and g1.
  abs(g2) is equal to that of g1.
      ****************************************/
{
  /*	double cas,cbs,ccs,bas,bbs,aas; */
  double hd,hn;
  /*	R_HKL h1,h2,h3; */
  R_HKL	t;
  
  /*
	h1.h = 1.0; h1.k = 0.0; h1.l = 0.0;
	h2.h = 0.0; h2.k = 1.0; h2.l = 0.0;
	h3.h = 0.0; h3.k = 0.0; h3.l = 1.0;
    
	cas = inner_product(h3, h1, bv);
	cbs = inner_product(h3, h2, bv);
	ccs = inner_product(h3, h3, bv);
	bas = inner_product(h2, h1, bv);
	bbs = inner_product(h2, h2, bv);
	aas = inner_product(h1, h1, bv);
    
	printf("$$ zolz_ort_base:\n");
	printf("$$ aas: %lf %lf\n", aas, bv.aas);
	printf("$$ bbs: %lf %lf\n", bbs, bv.bbs);
	printf("$$ ccs: %lf %lf\n", ccs, bv.ccs);
	printf("$$ cas: %lf %lf\n", cas, bv.acs);
	printf("$$ cbs: %lf %lf\n", cbs, bv.bcs);
	printf("$$ bas: %lf %lf\n", bas, bv.abs);
    
	t.h = bv.vc * (   u.v * (g1.h * cas + g1.k * cbs + g1.l * ccs)
					- u.w * (g1.h * bas + g1.k * bbs + g1.l * cbs) );
	t.k = bv.vc * (   u.w * (g1.h * aas + g1.k * bas + g1.l * cas)
					- u.u * (g1.h * cas + g1.k * cbs + g1.l * ccs) );
	t.l = bv.vc * (   u.u * (g1.h * bas + g1.k * bbs + g1.l * cbs)
					- u.v * (g1.h * aas + g1.k * bas + g1.l * cas) );
   */
  
  t.h = bv.vc * (
    u.v * (g1.h * bv.acs + g1.k * bv.bcs + g1.l * bv.ccs)
    - u.w * (g1.h * bv.abs + g1.k * bv.bbs + g1.l * bv.bcs)
    );
  t.k = bv.vc * (
    u.w * (g1.h * bv.aas + g1.k * bv.abs + g1.l * bv.acs)
    - u.u * (g1.h * bv.acs + g1.k * bv.bcs + g1.l * bv.ccs)
    );
  t.l = bv.vc * (
    u.u * (g1.h * bv.abs + g1.k * bv.bbs + g1.l * bv.bcs)
    - u.v * (g1.h * bv.aas + g1.k * bv.abs + g1.l * bv.acs)
    );
  
  hn = sqrt( sq_abs(g1, bv) );
  hd = sqrt( sq_abs(t,  bv) );
  
  if (hd < EPS) {
    printf("$Error in zolz_ort_base: g2 = 0\n");
    exit(EXIT_FAILURE);
  }
  
  t.h *= hn / hd; t.k *= hn / hd; t.l *= hn / hd;
  return t;
}


/* R_UVW zolz_ort_base_uvw(UVW u, R_HKL g1, LAT_PAR lp) */
R_UVW zolz_ort_base_uvw(UVW u, R_HKL g1, CELL_PAR bv)
     /*
      * orthogonal vector uvw
      */
{
  double cas,cbs,ccs,bas,bbs,aas;
  double ud,un;
  UVW u1,u2,u3;
  R_UVW t;
  
  u1.u = 1; u1.v = 0; u1.w = 0;   /* 100 */
  u2.u = 0; u2.v = 1; u2.w = 0;   /* 010 */
  u3.u = 0; u3.v = 0; u3.w = 1;   /* 001 */
  cas = inner_product_uvw(u3, u1, bv);
  cbs = inner_product_uvw(u3, u2, bv);
  ccs = inner_product_uvw(u3, u3, bv);
  bas = inner_product_uvw(u2, u1, bv);
  bbs = inner_product_uvw(u2, u2, bv);
  aas = inner_product_uvw(u1, u1, bv);
  t.u = (
    ((double)u.u * bas + (double)u.v * bbs + (double)u.w * cbs) * g1.l
    - ((double)u.u * cas + (double)u.v * cbs + (double)u.w * ccs) * g1.k
    ) / bv.vc;
  t.v = (
    ((double)u.u * cas + (double)u.v * cbs + (double)u.w * ccs) * g1.h
    - ((double)u.u * aas + (double)u.v * bas + (double)u.w * cas) * g1.l
    ) / bv.vc;
  t.w = (
    ((double)u.u * aas + (double)u.v * bas + (double)u.w * cas) * g1.k
    - ((double)u.u * bas + (double)u.v * bbs + (double)u.w * cbs) * g1.h
    ) / bv.vc;
  un = 1.0 / sqrt( sq_abs(g1, bv) );
  ud = abs_r_uvw(t, bv);
  if (ud < EPS) {
    printf("$Error in zolz_ort_base_uvw: g2 = 0\n");
    exit(EXIT_FAILURE);
  }
  t.u *= un / ud;	t.v *= un / ud; t.w *= un / ud;
  return t;
}


void zolz_ort_base_two_uvw(UVW u, R_HKL g1, CELL_PAR bv, R_UVW *u1, R_UVW *u2)
{
  /*
   * calc. of two real-space base-vectors u1 and u2
   *  from u and g1.
   */
  R_HKL g2;
  
  *u2 = zolz_ort_base_uvw(u, g1, bv);
  g2  = zolz_ort_base(u, g1, bv);
  *u1 = zolz_ort_base_uvw(u, g2, bv);
  u1->u = - u1->u;	u1->v = - u1->v;	u1->w = - u1->w;
}


R_HKL proj_to_zolz(
  HKL g, UVW u, R_HKL zolzbase1, CELL_PAR bv,
  double *coef_x, double *coef_y
  )
     /*
      * projection of reciprocal vector g to ZOLZ plane 
      */
{
  R_HKL h, proj_g, zolzbase2;
  double abs_base1,tx,ty;
  
  h.h = (double) g.h;	h.k = (double) g.k;	h.l = (double) g.l;
  
  zolzbase2 = zolz_ort_base(u, zolzbase1, bv);
  /* zolzbase2 is orthogonal to zolzbase1, and has the same norm. */
  abs_base1 = sq_abs(zolzbase1,bv);
  tx = inner_product(h, zolzbase1, bv) / abs_base1;
  ty = inner_product(h, zolzbase2, bv) / abs_base1;
  
  proj_g.h = zolzbase1.h * tx + zolzbase2.h * ty;
  proj_g.k = zolzbase1.k * tx + zolzbase2.k * ty;
  proj_g.l = zolzbase1.l * tx + zolzbase2.l * ty;
  
  *coef_x = tx; *coef_y = ty;
  
  return proj_g;	
}


R_HKL proj_to_zolz2(
  HKL g, UVW u, R_HKL zolzbase1, R_HKL zolzbase2, CELL_PAR bv,
  double *coef_x, double *coef_y
  )
     /*
      * projection of reciprocal vector g to ZOLZ plane 
      */
{
  R_HKL h, proj_g;
  double abs_base1,tx,ty;
  
  h.h = (double) g.h;	h.k = (double) g.k;	h.l = (double) g.l;
  
  abs_base1 = sq_abs(zolzbase1,bv);
  tx = inner_product(h, zolzbase1, bv) / abs_base1;
  ty = inner_product(h, zolzbase2, bv) / abs_base1;
  
  proj_g.h = zolzbase1.h * tx + zolzbase2.h * ty;
  proj_g.k = zolzbase1.k * tx + zolzbase2.k * ty;
  proj_g.l = zolzbase1.l * tx + zolzbase2.l * ty;
  
  *coef_x = tx; *coef_y = ty;
  
  return proj_g;	
}


R_HKL normalize_vec(R_HKL g, R_HKL h, CELL_PAR bv)
     /*
      * normalization of g (referring to the absolute value of h)
      */
{
  double  tg, th, coef;
  
  tg = sq_abs(g, bv);
  th = sq_abs(h, bv);
  
  if(tg < EPS) {
    g.h = 0.0; g.k = 0.0; g.l = 0.0;
    return g;
  }
  coef = sqrt( th / tg );
  g.h = coef * g.h; g.k = coef * g.k; g.l = coef * g.l;
  return g;
}


int extinction_rule(HKL g, char *lattice_type)
     /*
      * Extinction rule due to lattice type
      *   return 0     : possible
      *   return not 0 : forbidden
      *   11 Nov., 1995
      */
{
  int e;
  
  if ( lattice_type[0]=='P'){
    e = 0;
  }
  else if ( lattice_type[0]=='I' ){
    e = (g.h + g.k + g.l) % 2;
  }
  else if ( lattice_type[0]=='F' ){
    e = abs((g.h + g.k) % 2)
      + abs((g.k + g.l) % 2)
        + abs((g.l + g.h) % 2);
  }
  else if ( lattice_type[0]=='A' ){
    e = (g.k + g.l) % 2;
  }
  else if ( lattice_type[0]=='B' ){
    e = (g.h + g.l) % 2;
  }
  else if ( lattice_type[0]=='C' ){
    e = (g.h + g.k) % 2;
  }
  else if ( lattice_type[0]=='R' ){
    /* e = (- g.h + g.k + g.l) % 3;*/ /* for hexagonal cell */
    e = 0;	/* for primitive rhombohedral cell */
  }
  else{
    printf("$$ Lattice type %s is invalid", lattice_type);
    printf(" in extinction_rule() of lattice.c\n");
    exit(EXIT_FAILURE);
  }
  
  return (e);
}


UVW incident_direction(HKL h1, HKL h2)
     /*
      * Calculate the incidence vector defined
      *   from two reciprocal vectors
      */
{
  UVW u;
  
  u.u = h1.k * h2.l - h2.k * h1.l;
  u.v = h1.l * h2.h - h2.l * h1.h;
  u.w = h1.h * h2.k - h2.h * h1.k;
  return u;
}


#if 0
double chi(double accel_kvol)
     /*
      * chi (angstrom^{-1})
      */
{
  double rev;
  rev = accel_kvol * 1000.0;
  return sqrt( rev * (1.0 + rev * 9.784754e-7) ) / 12.26426;
}
#endif


double wavelength(double accel_kvol)
     /*
      * wave length [angstrom]
      *   = h / sqrt( 2 * me * e * E * ( 1 + e * E / (2 * me * sq(c)) )
      * input:	 accelerating voltage (kV)
      */
{
  double rev;
  rev = accel_kvol * 1000.0;
  return 12.264263 / sqrt( rev * (1.0 + rev * 9.7847537e-7) );
}


double relativ_corr(double accel_kvol)
     /*
      * Relativistic correction of structure factor (Gamma)
      *  = 1 + e * E / ( me * sq(c) )
      */
{
  /* 	return 1.0 + (1000.0 * accel_kvol) / 5.1099908e+5; */
  return 1.0 + (1000.0 * accel_kvol) * 1.9569507e-06;
}


double evdash(double accel_kvol)
{
     /*
      * evdash : corrected accel. voltage 
      * for Vg / evdash
      * accelerating voltage (kV)
      */
  double rev;
  rev = accel_kvol * 1000.0;
  return rev * ( 1.0 + 0.5 * rev / 5.109990e+5 ) / ( 1.0 + rev / 5.109990e+5 );
}
