/*
 * mbfit.c
 *
 * Programmed by Kenji Tsuda, IMRAM, Tohoku University, Sendai, JAPAN.
 *
 * Special thanks to
 *  Dr. K. Saitoh, Ms. N. Kanda, Mr. J. Shiono, Mr. K. Tanaka,
 *  Mr. K. Takagi, Mr. Y. Ogata
 *  and Prof. M. Tanaka.
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "mbconfig.h"
#include "fitting.h"
#include "mb_input.h"
#include "data_out.h"
#include "mb_sim.h"
#include "e_density.h"

#if defined(MPI_MODE)
#include <mpi.h>
int num_procs, proc_id;
#else
int num_procs=1, proc_id=0;
#endif

#define BUFLEN  256

DEBUG_INFO  dbg;
EPS_CTRL  epsilon;

void mb_fprintf(FILE *fp, char *fmt,...);
FILE *mb_fopen_chk_wmode(char *filename);
void mb_fclose(FILE *fp);
void mb_fflush(FILE *fp);
void read_eps_ctrl_params( FILE *fi );


void opening_title(void)
{
  mb_fprintf(stdout, "\n*****\n");
  mb_fprintf(
    stdout, "*****  MBFIT  [Many-Beam dynamical-simulations and least-squares FITting]\n"
    );
  mb_fprintf(stdout, "*****\n");
  mb_fprintf(
    stdout, "*****  Kenji Tsuda, IMRAM, Tohoku University, Sendai, JAPAN.\n"
    );

/*
  mb_fprintf(stdout, "*****\n");
  mb_fprintf(stdout, "*****    Special thanks to\n");
  mb_fprintf(stdout, "*****      Dr. K. Saitoh, Ms. N. Kanda, Mr. J. Shiono, Mr. K. Tanaka, \n");
  mb_fprintf(stdout, "*****      Mr. K. Takagi, Dr. K. Omoto, Mr. Y. Ogata,  Prof. M. Terauchi\n");
  mb_fprintf(stdout, "*****      and Prof. M. Tanaka.\n");
  mb_fprintf(stdout, "*****\n");
*/
  
#if defined(MPI_MODE)
  mb_fprintf(stdout, "*****  [MPICH version]\n");
#else
  mb_fprintf(stdout, "*****  [Single CPU version]\n");
#endif
}


void opening_title_log(FILE *fo_log)
{
  mb_fprintf(fo_log, "\n*****\n");
  mb_fprintf(
    fo_log, "*****  MBFIT  [Many-Beam dynamical-simulations and least-squares FITting]\n"
    );
  mb_fprintf(fo_log, "*****\n");
  mb_fprintf(
    fo_log, "*****  Kenji Tsuda, IMRAM, Tohoku University, Sendai, JAPAN.\n"
    );

/*
  mb_fprintf(fo_log, "*****\n");
  mb_fprintf(fo_log, "*****    Special thanks to\n");
  mb_fprintf(fo_log, "*****      Dr. K. Saitoh, Ms. N. Kanda, Mr. J. Shiono, Mr. K. Tanaka, \n");
  mb_fprintf(fo_log, "*****      Mr. K. Takagi, Dr. K. Omoto, Mr. Y. Ogata,  Prof. M. Terauchi\n");
  mb_fprintf(fo_log, "*****      and Prof. M. Tanaka.\n");
  mb_fprintf(fo_log, "*****\n");
*/
  
#if defined(MPI_MODE)
  mb_fprintf(fo_log, "*****  [MPICH version]\n");
#else
  mb_fprintf(fo_log, "*****  [Single CPU version]\n");
#endif
}


void ending_title(FILE *fo_log)
{
  mb_fprintf(fo_log, "\n");
  mb_fprintf(
    fo_log,
    " When you publish original results acquired with MBFIT,\n"
    );
  mb_fprintf( fo_log, " please cite the following paper:\n" );
  mb_fprintf(
    fo_log,
    " K. Tsuda and M. Tanaka: Acta Cryst., A55 (1999) 939-954.\n"
    );
  mb_fprintf( fo_log, "\nSee you again!!!\n" );
}


void mb_job( char *parfile, char *infile, FILE *fo_log )
{
  char buf[BUFLEN], job_mode[BUFLEN];
  FILE *fi;

  strcpy(dbg.infile, infile);  /* For debug */

  /* Open the parameter file */
  fi = fopen_chk_rmode(parfile);

  fget_1s(job_mode, fi);  /* Job mode selection */
  /* mb_fprintf(stdout, "$$ job_mode = %s\n", job_mode); */
  read_eps_ctrl_params(fi);

  if (strcmp(job_mode, "FIT_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    mb_fprintf(fo_log, "***** Non-linear least square fitting\n");
    mb_fprintf(stdout, "***** Non-linear least square fitting\n");
    mb_fprintf(fo_log, "*****\n");
    mb_fitting( fi, fo_log, infile );
  }
  else if	(strcmp(job_mode, "SIM_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    mb_fprintf(stdout, "*****\n");
    mb_fprintf(fo_log, "***** Many-beam dynamical simulation\n");
    mb_fprintf(stdout, "***** Many-beam dynamical simulation\n");
    mb_fprintf(fo_log, "*****\n");
    mb_fprintf(stdout, "*****\n");
    mb_simulation( fi, fo_log, infile );
  }
/*
  else if	(strcmp(job_mode, "TSCAN_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    mb_fprintf(fo_log, "***** Thickness scan for fit\n");
    mb_fprintf(fo_log, "*****\n");
    mb_thickness_scan( fi, infile );
  }
  else if	(strcmp(job_mode, "SHIFTSCAN_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    mb_fprintf(fo_log, "***** HG-shift scan for fit\n");
    mb_fprintf(fo_log, "*****\n");
    mb_shift_scan( fi, infile );
  }
 */
  /*
  else if	(strcmp(job_mode, "SPCELL_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    printf(" ## Many-beam dyn. sim. using the super-cell method ##\n");
    mb_fprintf(fo_log, "*****\n");
    mb_sim_spcell( fi, infile );
  }
  else if	(strcmp(job_mode, "COHERENT_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    printf(" ## Many-beam dyn. sim. of coherent CBED patterns ##\n");
    mb_fprintf(fo_log, "*****\n");
    mb_sim_coherent( fi, infile );
  }
   */
  else if (strcmp(job_mode, "DISPSURF_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    mb_fprintf(fo_log, "***** Calculation of dispersion surfaces\n");
    mb_fprintf(stdout, "***** Calculation of dispersion surfaces\n");
    mb_fprintf(fo_log, "*****\n");
    mb_dispersion_surface( fi, fo_log, infile );
  }
  else if (strcmp(job_mode, "BLOCH_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    mb_fprintf(fo_log, "***** Calculation of Bloch states\n");
    mb_fprintf(stdout, "***** Calculation of Bloch states\n");
    mb_fprintf(fo_log, "*****\n");
    mb_bloch_state( fi, fo_log, infile );
  }
  else if (strcmp(job_mode, "TFRINGE_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    mb_fprintf(fo_log, "***** Calculation of thickness fringes\n");
    mb_fprintf(stdout, "***** Calculation of thickness fringes\n");
    mb_fprintf(fo_log, "*****\n");
    mb_tfringe( fi, fo_log, infile );
  }
  else if (strcmp(job_mode, "E_DENSITY_PROJ_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    mb_fprintf(fo_log, "***** Calculation of projected electron density\n");
    mb_fprintf(stdout, "***** Calculation of projected electron density\n");
    mb_fprintf(fo_log, "*****\n");
    mb_electron_density_proj( fi, fo_log, infile );
  }
  else if (strcmp(job_mode,"E_DENSITY_3D_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    mb_fprintf(fo_log, "***** Calculation of 3D electron density\n");
    mb_fprintf(stdout, "***** Calculation of 3D electron density\n");
    mb_fprintf(fo_log, "*****\n");
    mb_electron_density_3d( fi, fo_log, infile );
  }
  /*
  else if (strcmp(job_mode,"XRD_E_DENSITY_PROJ_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    mb_fprintf(fo_log, "***** Calc. of proj. electron density (X-ray's f)\n");
    mb_fprintf(fo_log, "*****\n");
    mb_xrd_electron_density_proj( fi, infile );
  }
  else if (strcmp(job_mode,"NUC_C_DENSITY_PROJ_MODE") == 0){
    mb_fprintf(fo_log, "*****\n");
    mb_fprintf(fo_log, "***** Calc. of proj. nuclear charge density (Z-num)\n");
    mb_fprintf(fo_log, "*****\n");
    mb_nuclear_charge_density_proj( fi, infile );
  }
    */
  /*
  else if (strcmp(job_mode, "REFLEX") == 0){
    mb_fprintf(fo_log, "*****\n");
    printf(" ##### Reflection search #####\n");
    mb_fprintf(fo_log, "*****\n");
    mb_beam_search( fi, infile );
    }
   */
  /*
  else if (strcmp(job_mode, "STRCTF") == 0){
    mb_fprintf(fo_log, "*****\n");
    printf(" ##### Calc. of structure factors #####\n");
    mb_fprintf(fo_log, "*****\n");
    mb_calc_strctf( fi, infile );
  }
   */
  else {
    mb_fprintf(stdout, "*****\n");
    mb_fprintf(stdout, "***** Job mode %s is not supported. Abort.\n", job_mode);
    mb_fprintf(stdout, "*****\n");
    exit(EXIT_FAILURE);
  }
  return;
}


int main(int argc, char *argv[])
{
  char parfile[BUFLEN], infile[BUFLEN], log_outfile[BUFLEN];
  FILE *fi, *fo_log;

#if defined(MPI_MODE)
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &proc_id);
  MPI_Comm_size(MPI_COMM_WORLD, &num_procs);
#endif

  mb_fprintf(stdout, "\nWelcome to MBFIT!\n");
  read_ini_file();
  mb_fprintf(stdout, "\n");
  opening_title();
  
  if (argc == 1){
    mb_fprintf(stdout, "Parameter file (with extension) >>");
    fget_1s(parfile, stdin);
  }
  else if (argc == 2){
    strcpy(parfile, argv[1]);
  }
  else{
    mb_fprintf(stdout, "Usage: mbfit parameter_file (with extension)\n");
    exit(EXIT_FAILURE);
  }
  strcpy(infile, remove_extention(parfile));

  sprintf(log_outfile, "%s_log.txt", infile);
  fo_log = mb_fopen_chk_wmode(log_outfile);

  opening_title_log(fo_log);
  mb_fprintf(fo_log, "*****  Ver. 3.2a2 updated on 21 Jun. 2004.\n");
  mb_fprintf(stdout, "*****  Ver. 3.2a2 updated on 21 Jun. 2004.\n");
  mb_fprintf(fo_log, "*****\n");
  mb_fprintf(stdout, "*****\n");

  mb_fprintf(fo_log, "parameter_file>> %s\n", parfile);
  mb_fprintf(stdout, "parameter_file>> %s\n", parfile);

  mb_job(parfile, infile, fo_log);

  ending_title(fo_log);
  mb_fclose(fo_log);
  mb_fprintf(stdout, "\nSee you again!!\n" );
  mb_fprintf(stdout, "\n(MBFIT log file ==> %s)\n", log_outfile);

#if defined(MPI_MODE)
  MPI_Finalize();
#endif
  return EXIT_SUCCESS;
}
